
const
  TwoDigitLookupRaw : packed array[0..99] of array[1..2] of AnsiChar =
    ('00','01','02','03','04','05','06','07','08','09',
     '10','11','12','13','14','15','16','17','18','19',
     '20','21','22','23','24','25','26','27','28','29',
     '30','31','32','33','34','35','36','37','38','39',
     '40','41','42','43','44','45','46','47','48','49',
     '50','51','52','53','54','55','56','57','58','59',
     '60','61','62','63','64','65','66','67','68','69',
     '70','71','72','73','74','75','76','77','78','79',
     '80','81','82','83','84','85','86','87','88','89',
     '90','91','92','93','94','95','96','97','98','99');
var
  TwoDigitLookupW : packed array[0..99] of Word absolute TwoDigitLookupRaw;

var
  TwoDigitLookupLW : packed array[0..99] of LongWord;

{$If defined(Use_FastCodeFillChar) or defined(PatchSystemMove) or defined(USE_FAST_STRLEN)}
  {$D-} {Prevent Steppping into Move Code} //EH: moved after FastCode.inc is loaded to prevent debugging
  {$IFDEF VER170}
    {$DEFINE SSE2}
  {$ENDIF}

  {$IFDEF CONDITIONALEXPRESSIONS}
    {$IF CompilerVersion >= 18.0}
      {$DEFINE SSE2}
      {$WARN UNSAFE_CODE OFF}
      {$WARN UNSAFE_CAST OFF}
    {$IFEND}
  {$ENDIF}

{***** BEGIN LICENSE BLOCK *****
 FastcodeCPUID extraction
 Version: MPL 1.1

 The contents of this file are subject to the Mozilla Public License Version 1.1
 (the "License"); you may not use this file except in compliance with the
 License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 the specific language governing rights and limitations under the License.

 The Original Code is the FastCode CPUID code.

 The Initial Developer of the Original Code is
 Roelof Engelbrecht <roelof@cox-internet.com>. Portions created by
 the Initial Developer are Copyright (C) 2004 by the Initial Developer.
 All Rights Reserved.

 Contributor(s): Dennis Passmore <Dennis_Passmore@ ultimatesoftware.com>,
                 Dennis Christensen <marianndkc@home3.gvdnet.dk>,
                 Jouni Turunen <jouni.turunen@NOSPAM.iki.fi>.
                 John O'Harrow <john@elmcrest.demon.co.uk>

***** END LICENSE BLOCK *****

Version  Changes
-------  ------
 3.0.5   25-Feb-2008 : Added 2008 FastCode Targets
 3.0.4   20-Nov-2007 : Added SSE4A, SSE4.1, SSE4.2 and SSE5 detections. Added 2007 FastCode Targets
 3.0.3   20-Nov-2006 : Added SSSE3 Detection
 3.0.2   27 Apr 2006 : AMD X2 text changed from 'AMD_64_SSE3' to 'AMD_64X2'
 3.0.1   18 Apr 2006 : Bug in Yohan fctPMY target fixed, was incorrectly set to fctPMD
 3.0.0   27 Feb 2006 : Added new 2006 computed targets. Added Yonah and Presler
                       Removed Prescott, Banias, AMD XP                    (JT)

}

type
  TVendor = (cvUnknown, cvAMD, cvCentaur, cvCyrix, cvIntel,
    cvTransmeta, cvNexGen, cvRise, cvUMC, cvNSC, cvSiS);
  {Note: when changing TVendor, also change VendorStr array below}

  TInstructions =
    (isFPU, {80x87}
     isTSC, {RDTSC}
     isCX8, {CMPXCHG8B}
     isSEP, {SYSENTER/SYSEXIT}
     isCMOV, {CMOVcc, and if isFPU, FCMOVcc/FCOMI}
     isMMX, {MMX}
     isFXSR, {FXSAVE/FXRSTOR}
     isSSE, {SSE}
     isSSE2, {SSE2}
     isSSE3, {SSE3*}
     isSSSE3, {SSSE3}
     isSSE4A, {SSE4A}
     isSSE41, {SSE4.1}
     isSSE42, {SSE4.2}
     isSSE5,  {SSE5}
     isMONITOR, {MONITOR/MWAIT*}
     isCX16, {CMPXCHG16B*}
     isX64, {AMD AMD64* or Intel EM64T*}
     isExMMX, {MMX+ - AMD only}
     isEx3DNow, {3DNow!+ - AMD only}
     is3DNow); {3DNow! - AMD only}

  {Note: when changing TInstruction, also change InstructionSupportStr below
         * - instruction(s) not supported in Delphi 7 assembler}
  TInstructionSupport = set of TInstructions;

  TCPU = record
    Vendor: TVendor;
    Signature: Cardinal;
    EffFamily: Byte; {ExtendedFamily + Family}
    EffModel: Byte; {(ExtendedModel shl 4) + Model}
    EffModelBasic: Byte; {Just Model (not ExtendedModel shl 4) + Model)}
    CodeL1CacheSize, {KB or micro-ops for Pentium 4}
      DataL1CacheSize, {KB}
      L2CacheSize, {KB}
      L3CacheSize: Word; {KB}
    InstructionSupport: TInstructionSupport;
  end;

  TFastCodeTarget =
   (fctIA32,             {not specific to any CPU}
    fctIA32SizePenalty,  {not specific to any CPU, In library routines with size penalties used This target was called "fctRTLReplacement" earlier}
    fctMMX,              {not specific to any CPU, requires FPU, MMX and CMOV  "Old fctBlended target"}
    fctMMXSizePenalty,   {not specific to any CPU, requires FPU, MMX and CMOV  "Old fctBlended target" In library routines with size penalties used}
    fctSSE,              {not specific to any CPU, requires FPU, MMX, CMOV and SSE}
    fctSSE2,             {not specific to any CPU, requires FPU, MMX, CMOV, SSE, SSE2}
    fctSSE3,             {not specific to any CPU, requires FPU, MMX, CMOV, SSE, SSE2, SSE3}
    fctSSSE3,            {not specific to any CPU, requires FPU, MMX, CMOV, SSE, SSE2, SSE3, SSSE3}
    fctPascal,           {use Pascal routines in library}
    fctPascalSizePenalty,{use Pascal routines with size penalty in library}
    fctPMC, {Conroe}
    fctPMY, {Yonah}
    fctP4N, {Northwood}
    fctP4P, {Penryn: Core 2 with SSE4.1}
    fctAmd64_SSE4A, {AMD Phenom with SSE4A}
    fctAmd64_SSE3); {X2/Opteron/Athlon FX/Athlon 64 with SSE3}
  {Note: when changing TFastCodeTarget, also change FastCodeTargetStr array
         below}

const
  VendorStr: array[Low(TVendor)..High(TVendor)] of ShortString =
  ('Unknown', 'AMD', 'Centaur (VIA)', 'Cyrix', 'Intel', 'Transmeta',
    'NexGen', 'Rise', 'UMC', 'National Semiconductor', 'SiS');

  InstructionSupportStr:
  array[Low(TInstructions)..High(TInstructions)] of ShortString =
    ('FPU', 'TSC', 'CX8', 'SEP', 'CMOV', 'MMX', 'FXSR', 'SSE', 'SSE2', 'SSE3',
    'SSSE3', 'SSE4A', 'SSE41', 'SSE42', 'SSE5', 'MONITOR', 'CX16', 'X64', 'MMX+', '3DNow!+', '3DNow!');

  FastCodeTargetStr:
  array[Low(TFastCodeTarget)..High(TFastCodeTarget)] of ShortString =
    ('IA32', 'IA32_SizePenalty', 'MMX', 'MMX_SizePenalty', 'SSE',
     'SSE2', 'SSE3', 'SSSE3', 'Pascal', 'Pascal_SizePenalty',
     'Conroe', 'Yonah', 'Northwood', 'Penryn', 'Phenom', 'AMD_64X2');

var
  CPU: TCPU;
  FastCodeTarget: TFastCodeTarget;
{$IFEND}

{$IFDEF PatchSystemMove} //set in Zeos.inc
var
  Move : procedure(const Source; var Dest; Count : Integer); {Fastest Move}

{Procedures interfaced only for testing and validation purposes}
  procedure Move_JOH_IA32_10(const Source; var Dest; Count : Integer);
  procedure Move_JOH_MMX_10 (const Source; var Dest; Count : Integer);
  procedure Move_JOH_SSE_10 (const Source; var Dest; Count : Integer);
  procedure Move_JOH_SSE2_10(const Source; var Dest; Count : Integer);
  procedure Move_JOH_SSE3_10(const Source; var Dest; Count : Integer);
{$ENDIF PatchSystemMove} //set in Zeos.inc

{$IFDEF Use_FastCodeFillChar}
var
  FillChar: procedure (var Dest; count: Integer; Value: AnsiChar);
{$ENDIF Use_FastCodeFillChar}

var
  {$IFDEF FPC}
  StrLen: function(Str: PAnsiChar): {$IFDEF cpui386}LongInt{$ELSE}Int64{$ENDIF};
  {$ELSE}
  StrLen: function(const Str: PAnsiChar): Cardinal;
  {$ENDIF}

{$IFDEF USE_FAST_CHARPOS}
var
  CharPos: function(ch: char; const s: String): integer;
{$ENDIF}

function IntToStr(Value: Integer): String; overload;  //keep always this one @first pos becouse of the Ansi-Delphi faster BASM code (the int64 version call the 32bit within range)
function IntToStr(const Value: ShortInt): String; overload;
function IntToStr(Value: Byte; Const Negative: Boolean = False): String; overload;
function IntToStr(const Value: SmallInt): String; overload;
function IntToStr(Value: Word; Const Negative: Boolean = False): String; overload;
function IntToStr(Value: Cardinal; Const Negative: Boolean = False): String; overload;
function IntToStr({$IF not defined(Delphi) or defined(UNICODE)}const{$IFEND} Value: Int64): String; overload;
function IntToStr(Value: UInt64; Const Negative: Boolean = False): String; overload;

{ Integer convertion in Raw and Unicode Strings}
function IntToRaw(Value: Integer): RawByteString; overload;  //keep always this one @first pos becouse of the Ansi-Delphi faster BASM code (the int64 version call the 32bit within range)
function IntToRaw(const Value: ShortInt): RawByteString; overload;
function IntToRaw(Value: Byte; Const Negative: Boolean = False): RawByteString; overload;
function IntToRaw(const Value: SmallInt): RawByteString; overload;
function IntToRaw(Value: Word; Const Negative: Boolean = False): RawByteString; overload;
function IntToRaw(Value: Cardinal; Const Negative: Boolean = False): RawByteString; overload;
function IntToRaw({$IF not defined(Delphi) or defined(UNICODE)}const{$IFEND} Value: Int64): RawByteString; overload;
function IntToRaw(Value: UInt64; Const Negative: Boolean = False): RawByteString; overload;

function IntToUnicode(Value: Byte; Const Negative: Boolean = False): ZWideString; overload;
function IntToUnicode(const Value: ShortInt): ZWideString; overload;
function IntToUnicode(Value: Word; Const Negative: Boolean = False): ZWideString; overload;
function IntToUnicode(const Value: SmallInt): ZWideString; overload;
function IntToUnicode(Value: Cardinal; Const Negative: Boolean = False): ZWideString; overload;
function IntToUnicode(const Value: Integer): ZWideString; overload;
function IntToUnicode(const Value: Int64): ZWideString; overload;
function IntToUnicode(Value: UInt64; Const Negative: Boolean = False): ZWideString; overload;


function RawToInt(const Value: RawByteString): Integer;
function UnicodeToInt(const Value: ZWideString): Integer;
function RawToIntDef(const S: RawByteString; const Default: Integer) : Integer; overload;
function RawToIntDef(const S: PAnsiChar; const Default: Integer) : Integer; overload;
function UnicodeToIntDef(const S: ZWideString; const Default: Integer) : Integer; overload;
function UnicodeToIntDef(const S: PWideChar; const Default: Integer) : Integer; overload;
function RawToInt64Def(const S: RawByteString; const Default: Integer) : Int64; overload;
function RawToInt64Def(const S: PAnsiChar; const Default: Integer) : Int64; overload;
function RawToUInt64Def(const S: PAnsiChar; const Default: Cardinal) : UInt64; overload;
function RawToUInt64Def(const S: RawByteString; const Default: Cardinal) : UInt64; overload;
function UnicodeToInt64Def(const S: ZWideString; const Default: Integer) : Int64; overload;
function UnicodeToInt64Def(const S: PWideChar; const Default: Integer) : Int64; overload;

{ Float convertion in Raw and Unicode Format}
function RawToFloat(const s: RawByteString): Extended; overload;
function RawToFloat(const s: PAnsiChar): Extended; overload;
function RawToFloat(const s: RawByteString; const DecimalSep: AnsiChar): Extended; overload;
function RawToFloat(const s: PAnsiChar; const DecimalSep: AnsiChar): Extended; overload;
function RawToFloatDef(const s: RawByteString; const DecimalSep: AnsiChar; const Default: Extended): Extended; overload;
function RawToFloatDef(const s: PAnsiChar; const DecimalSep: AnsiChar; const Default: Extended): Extended; overload;
function ValRawExt(const s: RawByteString; const DecimalSep: AnsiChar; var code: Integer): Extended; overload;
function ValRawExt(const s: PAnsiChar; const DecimalSep: AnsiChar; var code: Integer): Extended; overload;
function ValRawInt(const s: RawByteString; var code: Integer): Integer;

function UnicodeToFloat(const s: ZWideString): Extended; overload;
function UnicodeToFloat(const s: PWideChar): Extended; overload;
function UnicodeToFloat(const s: ZWideString; const DecimalSep: WideChar): Extended; overload;
function UnicodeToFloat(const s: PWideChar; const DecimalSep: WideChar): Extended; overload;
function UnicodeToFloatDef(const s: ZWideString; const DecimalSep: WideChar; const Default: Extended): Extended;  overload;
function UnicodeToFloatDef(const s: PWideChar; const DecimalSep: WideChar; const Default: Extended): Extended; overload;
function ValUnicodeExt(const s: ZWideString; const DecimalSep: WideChar; var code: Integer): Extended; overload;
function ValUnicodeExt(const s: PWideChar; const DecimalSep: WideChar; var code: Integer): Extended; overload;

{Faster Floating functions ..}

{$IFDEF USE_FAST_TRUNC}
function Trunc(const X: Extended): Int64; overload;
function Trunc(const X: Double): Int64; overload;
function Trunc(const X: Single): Int64; overload;
{$ENDIF USE_FAST_TRUNC}
